/*
 * intf.c
 *
 * Copyright (C) 2002 Thomas Graf <tgr@kaosu.ch>
 */


#include "intf.h"


/*
 * must be implemented by the application using this module
 * is called by this module whenever a fatal error occurs
 */
extern void quit (const char *fmt, ...);


extern char cfg_fill_char;
extern char cfg_bg_char;
extern int cfg_hide_dummy_interfaces;
extern char *allowed_intf[255];
extern char *denied_intf[255];



intf_t  intf[MAX_IF];


#define SECOND     1.0f
#define MINUTE     60.0f
#define HOUR       3600.0f
#define TABLE_COLS 60


#define UPDATE_TS(tv) {                     \
    gettimeofday(&(tv), NULL);              \
}

#define COPY_TS(tv1,tv2) {                  \
    (tv1)->tv_sec = (tv2)->tv_sec;          \
    (tv1)->tv_usec = (tv2)->tv_usec;        \
}
    

#define INTF_MARK_ALL_REMOVE() {           \
    int i;                                 \
    for (i=0; i < MAX_IF; i++) {           \
        if (intf[i].name[0])               \
            intf[i].to_remove = 1;         \
    }                                      \
}

#define INTF_MARK_ALL_NOT_UPDATED() {      \
    int i;                                 \
    for (i=0; i < MAX_IF; i++) {           \
        if (intf[i].name[0])               \
            intf[i].already_updated = 0;   \
    }                                      \
}

#define INTF_REMOVED_UNUSED() {             \
    int i;                                  \
    for (i=0; i < MAX_IF; i++)              \
        if (intf[i].to_remove)              \
            memset(&intf[i], 0,             \
                sizeof(struct intf_s));     \
}


static void p_intf_calc_all_rate(int i);
static void p_intf_calc_rate(rate_t * rate, struct timeval *tv);
static void p_intf_calc_graphic(struct intf_table_s *table);
static void p_update_table(struct intf_table_s *table, float needed_diff,
        rate_t *rx, rate_t *tx);
static void p_intf_draw_graphic(struct intf_table_s *table, int g_col,
        int t_col, unsigned int rx_half_step, unsigned int tx_half_step);
static int p_is_intf_allowed(const char *name);
static int p_intf_match_mask(const char *mask, const char *str);
static int p_intf_search_intf(const char *name);


void
intf_init(void)
{
    memset(&intf, 0, sizeof(intf));
}


static void
p_update_table(struct intf_table_s *table, float needed_diff, rate_t *rx,
        rate_t *tx)
{
    float diff = p_diff_now(&table->last_calculation);
    
    if (diff >= needed_diff)
    {
        UPDATE_TS(table->last_calculation);

        if (!table->rx_cnt_before)
            table->rx_cnt_before = rx->cnt;

        table->rx_rate[table->index] = (rx->cnt - table->rx_cnt_before);
        table->rx_cnt_before = rx->cnt;
        table->rx_rate[table->index] /= diff;


        if (!table->tx_cnt_before)
            table->tx_cnt_before = tx->cnt;

        table->tx_rate[table->index] = (tx->cnt - table->tx_cnt_before);
        table->tx_cnt_before = tx->cnt;
        table->tx_rate[table->index] /= diff;

        p_intf_calc_graphic(table);

        if (table->index == TABLE_COLS - 1)
            table->index = 0;
        else
            table->index++;
    }
}


float
p_diff(struct timeval *t1, struct timeval *t2)
{
    float            diff;

    /* calculate delta to last calculation */
    long sec  = (t2->tv_sec - t1->tv_sec);
    long usec = (t2->tv_usec - t1->tv_usec);

    /*
     * e.g.
     *
     * t1 = 1 sec  200000 usec
     * t2 = 2 sec  100000 usec
     *
     * the delta must be 0 sec 900000 usec
     */
    if (usec < 0)
    {
        sec--;
        usec += 1000000;
    }

    /*
     * make a float out of it <sec>.<usec>
     */
    diff = (float) sec;
    diff += ((float) usec) / 1000000.0f;

    return diff;
}


float
p_diff_now(struct timeval *t1)
{
    struct timeval now;

    UPDATE_TS(now);

    return p_diff(t1, &now);
}


void
intf_readfile(const char *path)
{

    FILE *       fd;
    char         buf[512];
    char *       p;
    char *       s;
    int          inum         = 0;
    int          i;
    int          w;

    INTF_MARK_ALL_REMOVE();
    INTF_MARK_ALL_NOT_UPDATED();

    if (!(fd = fopen(path, "r")))
        quit("Unable to open file %s: %s", path, strerror(errno));

    /*
     * ignore header lines
     */
    fgets(buf, sizeof(buf), fd);
    fgets(buf, sizeof(buf), fd);

    /*
     * /proc/net/dev example:
     *
     * Inter-|   Receive                                                |  Transmit
     * face |bytes    packets errs drop fifo frame compressed multicast|bytes    packets errs drop fifo colls carrier compressed
     *     lo:  450444    5419    0    0    0     0          0         0   450444    5419    0    0    0     0       0          0
     *   eth0:15019201   48550    0    0    0     0          0         0  7288864   62412    0    0    0     0       0          0
     */

    for (; fgets(buf, sizeof(buf), fd) && inum < (MAX_IF - 1);)
    {
        if (buf[0] == '\r' || buf[0] == '\n')
            continue;

        /*
         *        ^  eth0: ...
         * look for -----^
         */
        if (!(p = strchr(buf, ':')))
            continue;

        /*
         * replace ':' with \0
         * point s to '   15019201   48550 ....'
         */
        *p = '\0';
        s = p+1;

        /* ignore all precending spaces, p points now to 'eth0' */
        for (p = &buf[0]; *p == ' '; p++);

        if ((i = p_intf_search_intf(p)) < 0)
            continue;

        /* this interface is still up, don't remove it */
        intf[i].to_remove = 0;
        intf[i].already_updated = 1;

        w = sscanf(s, "%llu %llu %llu %llu %llu %llu %llu %llu %llu %llu "
                "%llu %llu %llu %llu %llu %llu\n",
            &intf[i].rx_bytes.cnt,
            &intf[i].rx_packets.cnt,
            &intf[i].rx_errors,
            &intf[i].rx_drop,
            &intf[i].rx_fifo,
            &intf[i].rx_frame,
            &intf[i].rx_compressed,
            &intf[i].rx_multicast,
            &intf[i].tx_bytes.cnt,
            &intf[i].tx_packets.cnt,
            &intf[i].tx_errors,
            &intf[i].tx_drop,
            &intf[i].tx_fifo,
            &intf[i].tx_frame,
            &intf[i].tx_compressed,
            &intf[i].tx_multicast);

        if (w != 16)
            continue;

        UPDATE_TS(intf[i].read_ts);

        p_intf_calc_all_rate(i);

        p_update_table(&intf[i].s, SECOND,&intf[i].rx_bytes, &intf[i].tx_bytes);
        p_update_table(&intf[i].m, MINUTE,&intf[i].rx_bytes, &intf[i].tx_bytes);
        p_update_table(&intf[i].h, HOUR,  &intf[i].rx_bytes, &intf[i].tx_bytes);

        inum++;
    }

    fclose(fd);

    INTF_REMOVED_UNUSED();
}

void
intf_read_netstat(const char *netstat_cmd)
{
    FILE *       fd;
    char         buf[512];
    int          inum      = 0;
    char *       p;
    char *       s;
    int          i;
    int          w;

    INTF_MARK_ALL_REMOVE();
    INTF_MARK_ALL_NOT_UPDATED();

    fd = popen(netstat_cmd, "r");

    if (!fd)
        quit("popen(netstat...) failed: %s\n", strerror(errno));

    for (; fgets(buf, sizeof(buf), fd) && inum < (MAX_IF - 1);)
    {

        if (buf[0] == '\n' || buf[0] == '\r')
            continue;

        if (!strncmp(buf, "Name", 4) || !strncmp(buf, "Kernel", 6) || !strncmp(buf, "Iface", 5))
            continue;

        /*
         *       >hme0 <
         * look for --^
         */
        if (!(p = strchr(buf, ' ')))
            continue;

        /*
         * replace ' ' with \0
         */
        *p = '\0';
        s = p+1;

        /* ignore all precending spaces, p points now to 'eth0' */
        for (p = &buf[0]; *p == ' ' || *p == '\t'; p++);

        if ((i = p_intf_search_intf(p)) < 0)
            continue;

        /* this interface is still up, don't remove it */
        intf[i].to_remove = 0;
        intf[i].already_updated = 1;

#if defined SYS_LINUX
        w = sscanf(s, "%*s %*s %llu %llu %llu %*d %llu %llu %llu",
                &intf[i].rx_packets.cnt,
                &intf[i].rx_errors,
                &intf[i].rx_drop,
                &intf[i].tx_packets.cnt,
                &intf[i].tx_errors,
                &intf[i].tx_drop);

        if (w != 4)
            continue;
#else
        w = sscanf(s, "%*s %*s %*s %llu %llu %llu %llu %llu",
                &intf[i].rx_packets.cnt,
                &intf[i].rx_errors,
                &intf[i].tx_packets.cnt,
                &intf[i].tx_errors,
                &intf[i].tx_frame);

        if (w != 5)
            continue;
#endif
        UPDATE_TS(intf[i].read_ts);

        p_intf_calc_all_rate(i);

        p_update_table(&intf[i].s, SECOND,&intf[i].rx_bytes, &intf[i].tx_bytes);
        p_update_table(&intf[i].m, MINUTE,&intf[i].rx_bytes, &intf[i].tx_bytes);
        p_update_table(&intf[i].h, HOUR,  &intf[i].rx_bytes, &intf[i].tx_bytes);

        inum++;
    }

    pclose(fd);

    INTF_REMOVED_UNUSED();
}

#if defined HAVE_KSTAT && defined SYS_SUNOS
void
intf_read_kstat(void)
{
    kstat_ctl_t * kc;
    kstat_t * kst;
    int inum = 0;

    INTF_MARK_ALL_REMOVE();
    INTF_MARK_ALL_NOT_UPDATED();

    if (!(kc = kstat_open()))
        quit("kstat_open() failed");

    if ((kst = kstat_lookup(kc, NULL, -1, NULL)))
    {
        for (; kst && inum < (MAX_IF -1); kst = kst->ks_next)
        {
            int i;
            kstat_named_t *kn;

            if (strcmp(kst->ks_class, "net") != 0)
                continue;

            if (kstat_read(kc, kst, NULL) < 0)
                continue;

            if (!strcmp(kst->ks_name, "zero_copy"))
                continue;

            if ((i = p_intf_search_intf(kst->ks_name)) < 0)
                continue;

            /* this interface is still up, don't remove it */
            intf[i].to_remove = 0;
            intf[i].already_updated = 1;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "rbytes64")))
                intf[i].rx_bytes.cnt = kn->value.ui64;
            else
                if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "rbytes")))
                    intf[i].rx_bytes.cnt = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "ipackets64")))
                intf[i].rx_packets.cnt = kn->value.ui64;
            else
                if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "ipackets")))
                    intf[i].rx_packets.cnt = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "ierrors")))
                intf[i].rx_errors = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "multircv64")))
                intf[i].rx_multicast = kn->value.ui64;
            else
                if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "multircv")))
                    intf[i].rx_multicast = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "brdcstrcv")))
                intf[i].rx_broadcast = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "obytes64")))
                intf[i].tx_bytes.cnt = kn->value.ui64;
            else
                if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "obytes")))
                    intf[i].tx_bytes.cnt = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "opackets64")))
                intf[i].tx_packets.cnt = kn->value.ui64;
            else
                if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "opackets")))
                    intf[i].tx_packets.cnt = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "oerrors")))
                intf[i].tx_errors = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "multixmt64")))
                intf[i].rx_multicast = kn->value.ui64;
            else
                if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "multixmt")))
                    intf[i].rx_multicast = kn->value.ui32;

            if ((kn = (kstat_named_t *) kstat_data_lookup(kst, "brdcstrcv")))
                intf[i].rx_broadcast = kn->value.ui32;

            UPDATE_TS(intf[i].read_ts);

            p_intf_calc_all_rate(i);

            p_update_table(&intf[i].s, SECOND,&intf[i].rx_bytes, &intf[i].tx_bytes);
            p_update_table(&intf[i].m, MINUTE,&intf[i].rx_bytes, &intf[i].tx_bytes);
            p_update_table(&intf[i].h, HOUR,  &intf[i].rx_bytes, &intf[i].tx_bytes);

            inum++;
        }
    }

    kstat_close(kc);

    INTF_REMOVED_UNUSED();
}
#endif

#if defined SYS_BSD
void
intf_read_sysctl(void)
{
    int inum = 0;
    size_t n;
    int mib[] = {CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0};
    char *buf, *next, *lim;

    INTF_MARK_ALL_REMOVE();
    INTF_MARK_ALL_NOT_UPDATED();

    if (sysctl(mib, 6, NULL, &n, NULL, 0) < 0)
        quit("sysctl() failed");

    if (!(buf = calloc(1, n)))
        quit("Out of memory!\n");

    if (sysctl(mib, 6, buf, &n, NULL, 0) < 0)
        quit("sysctl() failed");

    lim = (buf + n);

    next = buf;
    while (next < lim && inum < (MAX_IF - 1))
    {
        struct if_msghdr *ifm, *nextifm;
        struct sockaddr_dl *sdl;

        ifm = (struct if_msghdr *) next;
        if (ifm->ifm_type != RTM_IFINFO)
            break;

        next += ifm->ifm_msglen;

        while (next < lim)
        {
            nextifm = (struct if_msghdr *) next;
            if (nextifm->ifm_type != RTM_NEWADDR)
                break;
            next += nextifm->ifm_msglen;
        }

        sdl = (struct sockaddr_dl *) (ifm + 1);


        if (ifm->ifm_flags & IFF_UP)
        {
            int i;

            if(sdl->sdl_family != AF_LINK)
                continue;

            if ((i = p_intf_search_intf(sdl->sdl_data)) < 0)
                continue;

            /* this interface is still up, don't remove it */
            intf[i].to_remove = 0;
            intf[i].already_updated = 1;

            intf[i].rx_bytes.cnt = ifm->ifm_data.ifi_ibytes;
            intf[i].rx_packets.cnt = ifm->ifm_data.ifi_ipackets;
            intf[i].rx_errors = ifm->ifm_data.ifi_ierrors;
            intf[i].tx_bytes.cnt = ifm->ifm_data.ifi_obytes;
            intf[i].tx_packets.cnt = ifm->ifm_data.ifi_opackets;
            intf[i].tx_errors = ifm->ifm_data.ifi_oerrors;

            UPDATE_TS(intf[i].read_ts);

            p_intf_calc_all_rate(i);

            p_update_table(&intf[i].s, SECOND,&intf[i].rx_bytes, &intf[i].tx_bytes);
            p_update_table(&intf[i].m, MINUTE,&intf[i].rx_bytes, &intf[i].tx_bytes);
            p_update_table(&intf[i].h, HOUR,  &intf[i].rx_bytes, &intf[i].tx_bytes);

            inum++;
        }

    }

    free(buf);

    INTF_REMOVED_UNUSED();
}
#endif


static void
p_intf_calc_rate(rate_t * rate, struct timeval *tv)
{
    float diff;

    if (!rate->cnt_before)
        rate->cnt_before = rate->cnt;

    /* diff between last calculation and read timestamp */
    diff = p_diff(&rate->last_calculation, tv);

    if (diff >= 1.0f)
    {
        if (rate->cnt)
        {
            rate->rate = (rate->cnt - rate->cnt_before);
            rate->rate /= diff;
            rate->cnt_before = rate->cnt;
        }

        COPY_TS(&rate->last_calculation, tv);
    }
}

static void
p_intf_calc_all_rate(int i)
{
    p_intf_calc_rate(&intf[i].rx_bytes, &intf[i].read_ts);
    p_intf_calc_rate(&intf[i].tx_bytes, &intf[i].read_ts);

    p_intf_calc_rate(&intf[i].rx_packets, &intf[i].read_ts);
    p_intf_calc_rate(&intf[i].tx_packets, &intf[i].read_ts);
}


static void
p_intf_calc_graphic(struct intf_table_s *table)
{

    /* reset table */
    {
        int n;

        memset(table->rx_graph, cfg_bg_char, sizeof(table->rx_graph));
        memset(table->tx_graph, cfg_bg_char, sizeof(table->tx_graph));

        for (n=0; n < 6; n++)
            table->rx_graph[n][60] = '\0';

        for (n=0; n < 6; n++)
            table->tx_graph[n][60] = '\0';
    }


    /*
     * - calculate max values
     * - calculate step value
     * - calculate scale
     */
    {
        int n;
        intf_cnt_t rx_max, tx_max;
        double rx_step, tx_step;


        rx_max = tx_max = 0;

        for (n=0; n < 60; n++)
        {
            if (rx_max < table->rx_rate[n])
                rx_max = table->rx_rate[n];

            if (tx_max < table->tx_rate[n])
                tx_max = table->tx_rate[n];
        }

        rx_step = rx_max / 6.0f;
        tx_step = tx_max / 6.0f;

        for (n=0; n < 6; n++)
        {
            table->rx_scale[n] = ((n + 1) * ceil(rx_step));
            table->tx_scale[n] = ((n + 1) * ceil(tx_step));
        }
    }

    /* fill 2 dimensional arrays */
    {
        int t_col, g_col;
        unsigned int rx_half_step = (table->rx_scale[0] >> 1);
        unsigned int tx_half_step = (table->tx_scale[0] >> 1);

        for (g_col=0, t_col = table->index; t_col >= 0; g_col++,t_col--)
            p_intf_draw_graphic(table, g_col, t_col, rx_half_step,
                    tx_half_step);
        for (t_col=59; g_col < 60; g_col++,t_col--)
            p_intf_draw_graphic(table, g_col, t_col, rx_half_step,
                    tx_half_step);
    }
}

static void
p_intf_draw_graphic(struct intf_table_s *table, int g_col, int t_col,
        unsigned int rx_half_step, unsigned int tx_half_step)
{
    intf_cnt_t rx, tx;

    if ((rx = table->rx_rate[t_col]))
    {
        int w;

        for (w = 0; w < 6; w++)
        {
            if (!w)
            {
                if (rx >= rx_half_step)
                    table->rx_graph[w][g_col] = cfg_fill_char;
            }
            else
            {
                if (rx >= (table->rx_scale[w-1] + rx_half_step))
                    table->rx_graph[w][g_col] = cfg_fill_char;
            }
        }
    }

    if ((tx = table->tx_rate[t_col]))
    {
        int w;

        for (w = 0; w < 6; w++)
        {
            if (!w)
            {
                if (tx >= tx_half_step)
                    table->tx_graph[w][g_col] = cfg_fill_char;
            }
            else
            {
                if (tx >= (table->tx_scale[w-1] + tx_half_step))
                    table->tx_graph[w][g_col] = cfg_fill_char;
            }
        }
    }
}

static int
p_is_intf_allowed(const char *name)
{
    int n;

    if (!allowed_intf[0] && !denied_intf[0])
        return 1;
    else
    {
        if (!allowed_intf[0])
        {
            for (n=0; n < 255 && denied_intf[n]; n++)
                if (p_intf_match_mask(denied_intf[n], name))
                    return 0;

            return 1;
        }
        else
        {
            for (n=0; n < 255 && denied_intf[n]; n++)
            {
                if (p_intf_match_mask(denied_intf[n], name))
                    return 0;
            }

            for (n=0; n < 255 && allowed_intf[n]; n++)
            {
                if (p_intf_match_mask(allowed_intf[n], name))
                    return 1;
            }

            return 0;
        }
    }
}


/* this is code from my ircd project */
static int
p_intf_match_mask(const char *mask, const char *str)
{
    int i; /* index for mask */
    int n; /* index for str */
    char c;

    if (!mask || !str)
        return 0;

    for (i=0,n=0; mask[i] != '\0'; i++,n++) {

        if (mask[i] == '*') {
            c = tolower(mask[i+1]);

            if (c == '\0') {
                /*
                 * blah* will match everything that begins with blah
                 */
                return 1;
            }

            /*
             * nananana
             */
            if (c == '+')
                c = ' ';

            /*
             * look for c in str, c is the character after the
             * wildcard in mask
             */
            for (; tolower(str[n]) != c; n++) {

                if (str[n] == '\0') {
                    /*
                     * str ends but the required character after *
                     * was not found -> mismatch
                     */
                    return 0;
                }
            }

            n--;
        } else if (mask[i] == '+') {
            if (str[n] !=  ' ')
                return 0;

        /*
         * mask[i] is not a special character, thus both, mask
         * and str, must match at the current index
         */
        } else if (tolower(mask[i]) != tolower(str[n])) {
            return 0;
        }
    }

    /*
     * mask ends, if str ends as well -> match
     */
    if (str[n] == '\0')
        return 1;
    else
        return 0;
}


static int
p_intf_search_intf(const char *name)
{
    int i;

    /* check if this interface is already known */
    for (i=0; i < MAX_IF; i++)
    {
        if (intf[i].name)
        {
            if (!strcmp(name, intf[i].name))
            {
                if (!intf[i].already_updated)
                    return i;
                else
                    return -1;
            }
        }
    }

	if (cfg_hide_dummy_interfaces)
	{
		if (!strcmp(name, "teql0") ||
			!strcmp(name, "shaper0") ||
			!strcmp(name, "bond0") ||
			!strcmp(name, "tunl0") ||
			!strcmp(name, "gre0") ||
			!strcmp(name, "sit0") ||
			!strcmp(name, "irlan0"))
		{
			return -1;
		}
	}


    if (!p_is_intf_allowed(name))
        return -1;


    /* nada, look for a free slot to save this interface */
    for (i=0; i < MAX_IF; i++)
    {
        if (!intf[i].name[0])
        {
            /*
             * hai! found one, store interface name into it and
             * update time
             */
            strncpy(intf[i].name, name, sizeof(intf[i].name) - 1);
            UPDATE_TS(intf[i].m.last_calculation);
            UPDATE_TS(intf[i].h.last_calculation);
            UPDATE_TS(intf[i].s.last_calculation);
            UPDATE_TS(intf[i].rx_bytes.last_calculation);
            UPDATE_TS(intf[i].tx_bytes.last_calculation);
            UPDATE_TS(intf[i].rx_packets.last_calculation);
            UPDATE_TS(intf[i].tx_packets.last_calculation);
            return i;
        }
    }

    return -1;
}
